#!/bin/bash
########################################################################
# build_sphinx.sh
#
# Build sphinx and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_08_03.
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: ${*}\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
    cat <<EOF
USAGE: ${prog} <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
       ${prog} -h|--help [<product-dir>]

  Build ${package} and package for UPS.

OPTIONS

  -h|--help

    Help. Supply <product-dir> for specific, detailed information on
    invoking ${prog}.

  <ssi-msg-config-options>

    Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    was unpacked. When specified with -h|--help, this enables us to
    provide more specific help about invoking ${prog}.

  <qualifier-args>
  <maketar>

    Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    build_script_usage
    exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "${@}"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="${arg}"; break
  done
  if [ -z "${product_dir}" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> ${product_dir}\n$(build_script_usage)"
    exit 2
  elif ! { [ -d "${product_dir}/.upsfiles" ] && [ -w "${product_dir}" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
      "${product_dir}\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "${product_dir}" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: ${product_dir}\n" \
    "$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="${product_dir}/ssibuildshims/${ssibuildshims_version}"
  [ -f "${ssibuildshims_topdir}/bin/ssi_functions" ] && \
    source "${ssibuildshims_topdir}/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "${product_dir}" || ssi_die "unable to setup UPS"
    setup ssibuildshims ${ssibuildshims_version} -z "${product_dir}" || \
    ssi_die "unable to find ssibuildshims ${ssibuildshims_version} in
${product_dir} and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "${product_dir}" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  $trace verify_in_filesystem "${@}" -x "${pkgdir}/bin/sphinx-build" && \
    (( ${#pvers_done[@]} > 0 )) || return
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "${maketar}" = "tar" ]; then
    # Customize mbd_opts. *C*
    local mbd_opts=()
    $trace is_built "$@" && \
      $trace make_basic_distribution "${@}" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Package-specific functions.
########################################################################

# Set up the required Python environment (native or UPS).
init_Python() {
  local pver=${1:+"$1"} # Empty for native.

  # Save the current value of UPS_OVERRIDE and remove any -B
  # options. Take account of possible argument bundling.
  local ups_override="$UPS_OVERRIDE"
  while [[ "$UPS_OVERRIDE" =~ ^(.*)(^|[[:space:]]+)-([A-Za-z0-9]*B[A-Za-z0-9]*) ]]; do
    local repl="${BASH_REMATCH[3]//B}"
    UPS_OVERRIDE="${BASH_REMATCH[1]}${repl:+${BASH_REMATCH[2]}-${repl}}${UPS_OVERRIDE#$BASH_REMATCH}"
  done
  if [ -n "$pver" ]; then # UPS.
    ver_array "$pver" pvArray
    setup_python \
      --latest={--ge,${pvArray[0]}.${pvArray[1]:-0},--lt,${pvArray[0]}.$(( ${pvArray[1]:-0} + 1 ))} || \
      ssi_die "unable to setup Python $pver"
  else # Native.
    if [ -n "$SETUP_PYTHON" ]; then
      unsetup python 2>/dev/null
    fi
    if [ -n "$SETUP_SQLITE" ]; then
      unsetup sqlite 2>/dev/null
    fi
    [ -z "$SETUP_PYTHON$SETUP_SQLITE" ] || \
      ssi_die "unable to clear UPS Python from environment"
  fi
  python_vars python3 || \
    ssi_die "unable to configure Python environment for python3 ${pver:+\"$pver \"}($(which python3 2>&1))"
  [ -z "$pver" ] || [ "$pythonMdotm" = "$pver" ] || \
    ssi_die "minor version of found Python ($pythonMdotm)${python:+ from $python} does not match required $pver"
  [ -z "$ups_override" ] || export UPS_OVERRIDE="$ups_override"
  [ "$ups_override" = "$UPS_OVERRIDE" ] || \
    ssi_die "unable to restore value of UPS_OVERRIDE in environment"
}

# Build Sphinx with our chosen Python.
build_with_Python() {
  local OPT OPTIND OPTARG native=
  while getopts :n OPT; do
    case $OPT in
      n) native=1;;
      *) break
    esac
  done
  shift $(( OPTIND - 1 ))
  OPTIND=1

  # Make sure we're using the intended Python.
  init_Python ${1:+"$1"} || return

  ##################
  # Make sure we have a (temporary, private) up-to-date pip.
  ssi_info "ensuring our own (private) up-to-date pip"
  pip_install_tree="$pip_install_top/python$pythonMdotm"
  mkdir -p "$pip_install_tree" || ssi_die "unable to prepare temporary Python area for pip"
  $trace env PATH=$(dropit -sfe "$pip_install_tree/bin") \
    PYTHONPATH=$(dropit -sfe -p "$PYTHONPATH" \
    "$pip_install_tree/lib/python$pythonMdotm/site-packages") \
    pip3 install -I --no-cache-dir --upgrade-strategy only-if-needed \
    --prefix "$pip_install_tree" -U pip pip-tools "setuptools<58.0" || \
    ssi_die "unable to ensure suitable versions of pip, pip-tools, and setuptools"

  ssi_info "packaging ${package} ${pkgver} for Python ${pythonMdotm}"

  # Set up Sphinx's UPS environment.
  $trace setup -B ${package} ${pkgver} \
    -z "${pkgtop}/fakedb" -z "${product_dir}" -z "${PRODUCTS}" || \
    ssi_die "unable to set up environment for $package $pkgver with Python $pythonMdotm"

  ##################
  # Install everything at once.

  # We need -I in all cases here:
  #
  # * For installs with system Python, don't want to be rendered
  #   non-portable by the presence of some non-default packages.
  #
  # * For installs with UPS Python, don't want to remove old versions of
  #   packages in UPS Python tree if we have access to them.
  local install_opts=(
    -I # See above.
    --disable-pip-version-check # Already ensured latest.
    --no-cache-dir # Avoid confusion.
    --prefix "${pkgdir}"
    --upgrade-strategy only-if-needed # Don't be greedy.
    --use-pep517 # Produce "standard" packages.
  )

  local requirements_ver=requirements-Python-$pythonMdotm.txt
  [ -f "$patchdir/$requirements_ver" ] && $trace cp "$patchdir/$requirements_ver" .
  # Install with a suitable environment.
  ( export PATH=$(dropit -sfe "$pip_install_tree/bin")
    export PYTHONPATH=$(dropit -sfe -p "$PYTHONPATH" \
    "$pkgdir/"{lib,lib64}"/python$pythonMdotm/site-packages" \
    "$pip_install_tree/lib/python$pythonMdotm/site-packages")
    { [ -r "$requirements_ver" ]  || \
        { $trace pip-compile --header --annotate $patchdir/requirements.in \
                 --resolver=backtracking \
                 --cache-dir pip-tools-cache-$pythonMdotm \
                 -o "$requirements_ver" &&
            $trace sed -Ei'' -e 's&^([^@[:space:]]+)[[:space:]]*@ .*$&./\1&; t fixdash; b; :fixdash; s&-&_&g' "$requirements_ver"
        }
    } &&
      $trace pip3 install ${install_opts[*]:+"${install_opts[@]}"} \
             -r "$requirements_ver" ) ||
    ssi_die "pip3 install of $package $pkgver FAILED with Python $pythonMdotm"

  ##################
  # Make sure we can generate a project.
  ( TMP=`mktemp -d -t build_sphinx.sh.XXXXXX`
    trap "[ -d \"$TMP\" ] && rm -rf \"$TMP\" 2>/dev/null" EXIT
    export PYTHONPATH=$(dropit -sfe -p "$PYTHONPATH" \
      "$pkgdir/"{lib,lib64}"/python$pythonMdotm/site-packages")
    cd "$TMP" && \
      $trace sphinx-quickstart --project TEST --sep \
      --author $(whoami) --release x.y.z --language en . </dev/null || \
      ssi_die "simple quick-start failed with Python ${pythonMdotm}" ) || exit

  # Clean up Sphinx's environment (Python etc. handled elsewhere).
  unsetup ${package}
}

# Compile all the things to Python bytecode.
recompile_for() {
  init_Python ${1:+"$1"} || return
  ssi_info "recompiling postprocessed Python source in ${package} ${pkgver} for Python ${pythonMdotm}..."
  setup -B ${package} ${pkgver} -z "${product_dir}" -z "${PRODUCTS}" || \
    ssi_die "unable to set up environment for $package $pkgver with Python $pythonMdotm"
  # Note that we need to recompile urllb3 serially in order to avoid a
  # lockup on some systems (e.g. Jenkins SLF7).
  $trace ${python} -m compileall -x /site-packages/urllib3/ -j ${ncores:-1} \
    "$pkgdir/bin" "$pkgdir/"{lib,lib64}"/python$pythonMdotm/site-packages" && \
    $trace ${python} -m compileall -j 1 \
    "$pkgdir/"{lib,lib64}"/python$pythonMdotm/site-packages/urllib3" || \
    ssi_die "unable to recompile postprocessed Python source in ${package} ${pkgver} for Python ${pythonMdotm}"
}

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
thisdir=$(cd "${BASH_SOURCE%/*}" && pwd -P)
trace=maybe_xtrace # Could set to xtrace instead for unconditional.
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
. ${thisdir}/version_info.sh || \
  ssi_die "unable to load version info from ${thisdir}/version_info"
sphinx_python_minver=3.9
####################################

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "${@}"

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals \
  --parse --usage-file=- -- "$@" <<EOF || exit # *C*
ENVIRONMENT

  DEBUG_(CONFIG|BUILD|TEST|INSTALL)

    Verbose output from the corresponding step.

  DEBUG_PATCHES
  VERBOSE_PATCHES

    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.
EOF
##################

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.
$trace define_basics NO_QUAL IN_SOURCE TRACK_PROGRESS || exit # *C*
##################

####################################
# Declare ourselves privately (no automatic setup for this product).
$trace fake_declare_product "${ssi_args[@]}" -- || exit
####################################

####################################
# Build-only dependencies. *C*
#
# e.g.
#   $trace setup_cmake --latest={--ge,3.15} || exit
#      (version requirement necessary for cmake --install).
# AND/OR
#   $trace setup_ninja --optional && (( want_ninja = 1 ))
####################################

# Print active products.
echo
ups active
echo

# Possible short-circuit if we're already built and declared.
$trace maybe_make_distribution --silent

ssi_info "building ${package} for ${OS}${qualdir:+-${qualdir}} (flavor ${pkgflvr})"

# Ensure directories: pkgdir, sourcedir, blddir.
$trace init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

$trace cd "$sourcedir" \
  || ssi_die "unable to cd to source directory ${sourcedir}"

# Source tarballs and patches. See prepare_source -h / patch_source -h
# for options.
for python_pkg in sphinx breathe sphinx_design; do
  $trace eval prepare_source "${ssi_args[@]}" -P "$patchdir" \
         -d "$python_pkg" -e "--strip-components,1" \
         \""$tardir/\$${python_pkg}_pkgtarfile"\" ||
    exit
done

shopt -s nullglob

# Bail early if DEBUG_PATCHES is set.
[ -n "${DEBUG_PATCHES}" ] && exit 1
####################################

####################################
# Prepare to build. *C*
#
# Set configure / compile flags, environment variables, etc.

# Survey available Python UPS packages.
pvers=\
($(ups list -aK+ python | while read pkg ver flav; do
      ! ver_greater "$sphinx_python_minver" "$ver" && \
      eval [[ $ver =~ '^v([[:digit:]]+)(_([[:digit:]]+))?' ]] && \
        (( ${BASH_REMATCH[1]} == 3 )) && \
        echo ${BASH_REMATCH[1]}${BASH_REMATCH[3]:+".${BASH_REMATCH[3]}"}
    done | sort -uVr))

# Level of parallelism for bytecode recompile.
ncores=$(ncores 2>/dev/null) || ncores=1

####################################
# Configure, build, and install.
$trace cd "$blddir" || \
  ssi_die "unable to cd to build directory $blddir"

# Somewhere for temporary installations of pip:
pip_install_top="${WORKSPACE:-${TMPDIR:-/var/tmp/$USER.$$}}/PythonTmp"
trap "rm -rf \"$pip_install_top\" 2>/dev/null" EXIT

python_vars /usr/bin/python3 >/dev/null 2>&1
if [ -n "$pythonMdotm" ]; then
  if ! ver_greater "$sphinx_python_minver" "$pythonMdotm"; then     
    $trace build_with_Python -n && pvers_done+=("")
  else
     ssi_warn "Native Python ($python == $pythonMdotm) must be >= $sphinx_python_minver to support sphinx"
  fi
fi

for pver in "${pvers[@]}"; do
  $trace build_with_Python ${pver:+"${pver}"} && pvers_done+=("${pver}")
done

# Find .py and executable text files (a.k.a. Python script candidates)
# for postprocessing.
ssi_info "finding Python files and scripts for postprocessing..."
IFS=$'\n' read -r -d '' -a pfiles < \
  <(find "$pkgdir" -type f \( \( -name '*.py' -print \) -o \
  \( -executable -exec file -e compress -e elf -e tar --mime-type \{\} \; \) \) | \
  sed -Ene '/:/ ! p; s&:[[:space:]]+text/.*$&&p')

ssi_info "postprocessing Python files and scripts for relocatability..."

postprocess_python_files "${pfiles[@]}"|| \
  ssi_die "postprocessing of python files for $package $pkgver"

# Make sure the scripts can find everything they need.
find "$pkgdir/bin" -maxdepth 1 -type f | \
  xargs perl -wpi'' -e 'BEGIN { $done = 0; }
$done or
  m&^\s*(\#|(?:from|import)\s+__future__|$)& or
    (m&^\s*(if\s*\(|try:|from\b|import\b)& and
     print '"'"'
########################################################################
# Added by build_sphinx.sh
import os
import sys
import distutils.sysconfig
sys.path = [ os.path.join(os.environ["SPHINX_FQ_DIR"],
             l,
             "python{0}".format(distutils.sysconfig.get_python_version()),
             "site-packages") for l in ("lib64", "lib") ] + sys.path
########################################################################

'"'"' and
     $done = 1);
$done = 0 if eof;' || \
  ssi_die "unable to add relocatable sys.path information to executable Python scripts for $package $pkgver"

ssi_info "postprocessing complete"

# Add -c for current declaration if required.
$trace declare_product -x -- || exit

# Need to recompile fixed binaries.
for pver in "${pvers_done[@]}"; do
  $trace recompile_for ${pver}
done

ssi_info "finished packaging ${package} ${pkgver}"
##################

#################
# Verify installation and make distribution archive. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
$trace is_built && $trace ssi_build_completed && \
  ssi_info "${package} is installed at ${pkgdir}" && \
  $trace maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
